"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const React = __importStar(require("react"));
const RpcInterface_1 = require("./RpcInterface");
const AnimatedCoders_1 = require("./AnimatedCoders");
const schema_pb_1 = require("./schema_pb");
const StructCoders_1 = require("./StructCoders");
const DecodeElement_1 = require("./DecodeElement");
class InstanceCounter {
    constructor() {
        this.freed = [];
        this.current = 1;
    }
    next() {
        const value = this.freed.pop();
        if (value) {
            return value;
        }
        else {
            const id = this.current;
            this.current += 1;
            return id;
        }
    }
    free(id) {
        this.freed.push(id);
    }
}
const instanceCounter = new InstanceCounter();
const createRetainedInstanceContainer = () => {
    const clocks = new Map();
    const values = new Map();
    const nodes = new Map();
    return {
        getClock: n => {
            return clocks.get(n.getNodeid());
        },
        setClock: (n, clock) => {
            return clocks.set(n.getNodeid(), clock);
        },
        getValue: n => {
            return values.get(n.getNodeid());
        },
        setValue: (node, value) => {
            return values.set(node.getNodeid(), value);
        },
        getNode: n => {
            return nodes.get(n.getNodeid());
        },
        setNode: (n, value) => {
            nodes.set(n.getNodeid(), value);
        },
        clear: () => {
            nodes.clear();
            values.clear();
            clocks.clear();
        },
    };
};
const reactElementFromMessage = (message, localProps, rpc, retainedInstances) => {
    return message
        .getUpdate()
        .getElementList()
        .map(elem => DecodeElement_1.ofEncodedReactElement((messageId, args) => {
        rpc.rpcCall(messageId, args ? args : {});
    }, key => {
        return localProps[key];
    }, elem, retainedInstances));
};
const poorMansShallowCompare = (t1, t2) => {
    let changed = false;
    const keys = [...Object.keys(t1), ...Object.keys(t2)];
    keys.forEach(key => {
        if (t1[key] !== t2[key]) {
            changed = true;
        }
    });
    return changed;
};
const useRemoteUpdateState = (webSocket, name, inputProps, localProps, rpcInterface) => {
    let updateKind = undefined;
    const localPropsRef = React.useRef({});
    const memoizedWebSocket = React.useRef(undefined);
    const memoizedProps = React.useRef({});
    React.useEffect(() => {
        updateKind = { type: "RemoteRender" };
    }, [name]);
    React.useEffect(() => {
        const nextProps = inputProps(rpcInterface);
        if (poorMansShallowCompare(memoizedProps.current, nextProps)) {
            updateKind = { type: "RemoteRender" };
        }
        memoizedProps.current = nextProps;
    }, [inputProps]);
    React.useEffect(() => {
        localPropsRef.current = localProps;
        if (!updateKind) {
            updateKind = { type: "LocalReRender" };
        }
    }, [localProps]);
    React.useEffect(() => {
        const ws = webSocket();
        if (memoizedWebSocket.current !== ws) {
            updateKind = {
                type: "Reconnect",
                oldWs: memoizedWebSocket.current,
            };
            memoizedWebSocket.current = ws;
        }
    });
    return {
        updateKind: () => updateKind,
        computedProps: () => ({
            ws: memoizedWebSocket.current,
            localPropsRef: localPropsRef,
            inputProps: memoizedProps.current,
        }),
    };
};
const connect = (ws, instanceId, onMessage) => {
    const eventListener = (event) => {
        const message = schema_pb_1.ServerToClient.deserializeBinary(new Uint8Array(event.data));
        if (message.getRootid() === instanceId) {
            onMessage(message);
        }
    };
    ws.addEventListener("message", eventListener);
    return () => {
        if (ws.readyState === WebSocket.OPEN) {
            const msg = new schema_pb_1.ClientToServer();
            msg.setUnmount(new schema_pb_1.Unmount());
            msg.setRootid(instanceId);
            ws.send(msg.serializeBinary());
        }
        ws.removeEventListener("message", eventListener);
    };
};
const useInstanceId = () => {
    const instanceId = React.useRef(0);
    React.useEffect(() => {
        instanceId.current = instanceCounter.next();
        return () => {
            instanceCounter.free(instanceId.current);
        };
    }, []);
    return () => instanceId.current;
};
const sendRender = (ws, name, props, localProps, instanceId) => {
    const msg = new schema_pb_1.ClientToServer();
    const payload = new schema_pb_1.Render();
    payload.setName(name);
    payload.setProps(StructCoders_1.toStruct(props));
    payload.setLocalpropsList(Object.keys(localProps));
    msg.setRootid(instanceId);
    msg.setRender(payload);
    const renderMessage = msg.serializeBinary();
    if (ws.readyState === WebSocket.OPEN) {
        ws.send(renderMessage);
    }
    else {
        const handler = () => {
            ws.send(renderMessage);
            ws.removeEventListener("open", handler);
        };
        ws.addEventListener("open", handler);
    }
};
exports.useNars = (webSocket, name, inputProps, localProps) => {
    const instanceId = useInstanceId();
    const [renderedElement, updateRenderedElement] = React.useState("Loading");
    const rpcInterface = RpcInterface_1.useRpcInterface(instanceId, webSocket);
    const [retainedInstances] = React.useState(() => createRetainedInstanceContainer());
    const disconnect = React.useRef(undefined);
    // It internally stores the latest reference to webSocket
    // So that we can determine if the webSocket we're using
    // has changed.
    const remoteUpdateState = useRemoteUpdateState(webSocket, name, inputProps, localProps, rpcInterface);
    // A ref is used so that the asynchrnous callbacks from the server
    // use the latest localProps (it happens when the state updates on
    // the server but there were no state transitions on the client.
    const lastRenderMessage = React.useRef(undefined);
    React.useEffect(() => {
        const update = remoteUpdateState.updateKind();
        if (update) {
            const { ws, inputProps, localPropsRef } = remoteUpdateState.computedProps();
            switch (update.type) {
                case "RemoteRender":
                    sendRender(ws, name, inputProps, localPropsRef.current, instanceId());
                    break;
                case "Reconnect":
                    if (disconnect.current) {
                        disconnect.current();
                        retainedInstances.clear();
                    }
                    disconnect.current = connect(ws, instanceId(), message => {
                        if (message.hasError()) {
                            updateRenderedElement(currentValue => {
                                if (currentValue === "Loading") {
                                    return "Error";
                                }
                                else {
                                    return currentValue;
                                }
                            });
                        }
                        else if (message.hasUpdate()) {
                            updateRenderedElement({
                                type: "Rendered",
                                element: reactElementFromMessage(message, localPropsRef.current, rpcInterface, retainedInstances),
                            });
                            lastRenderMessage.current = message;
                        }
                        else if (message.hasAnimatedvalueupdate()) {
                            const update = message.getAnimatedvalueupdate();
                            const value = update.getValue();
                            const toValue = update.getTovalue();
                            AnimatedCoders_1.updateAnimatedValue(value, toValue, retainedInstances);
                        }
                        else if (message.hasRpccall()) {
                            const call = message.getRpccall();
                            const messageId = call.getMessageid();
                            const args = call.getArg();
                            rpcInterface.executeIncomingRpcCall(messageId, args);
                        }
                    });
                    sendRender(ws, name, inputProps, localPropsRef.current, instanceId());
                    break;
                case "LocalReRender":
                    if (lastRenderMessage.current) {
                        updateRenderedElement({
                            type: "Rendered",
                            element: reactElementFromMessage(lastRenderMessage.current, localPropsRef.current, rpcInterface, retainedInstances),
                        });
                    }
                    break;
            }
        }
    });
    React.useEffect(() => {
        // Disconnect on unmount
        return () => {
            if (disconnect.current) {
                disconnect.current();
            }
        };
    }, []);
    return renderedElement;
};
exports.RemoteComponent = (props) => {
    const ws = props.webSocket;
    const renderedElement = exports.useNars(typeof ws === "function" ? ws : () => ws, props.name, props.inputProps, props.localProps);
    if (renderedElement === "Error" && props.renderError) {
        return props.renderError();
    }
    else if (typeof renderedElement === "object" &&
        renderedElement.type === "Rendered") {
        return React.createElement(React.Fragment, {}, ...renderedElement.element);
    }
    else if (props.renderLoading) {
        return props.renderLoading();
    }
    return null;
};
exports.useWebSocket = (url, shouldReconnect) => {
    const [ref, setRef] = React.useState({ current: null });
    const removeErrorListenerRef = React.useRef();
    return () => {
        if (ref.current) {
            return ref.current;
        }
        else {
            const ws = new WebSocket(url);
            if (shouldReconnect) {
                const handler = () => {
                    setTimeout(function () {
                        setRef(x => (x.current === ws ? { current: null } : x));
                    }, 200);
                };
                ws.addEventListener("close", handler);
                removeErrorListenerRef.current = () => {
                    ws.removeEventListener("close", handler);
                };
            }
            else if (removeErrorListenerRef.current) {
                removeErrorListenerRef.current();
            }
            ref.current = ws;
            return ws;
        }
    };
};
